package main

import (
	"github.com/fatih/structtag"
	"go/token"
	"net/http"
)

import (
	"go/ast"
	"go/parser"
	"log"
	"strings"
)

var api API

func addStruct(a *ast.GenDecl) {
	md := manageComments(a.Doc)
	if md["API"] == "" {
		return
	}
	tp := APIType{
		Name:   "",
		Desc:   "",
		Fields: make(map[string]*APIField),
		Col:    md["COL"],
	}
	tp.Name = a.Specs[0].(*ast.TypeSpec).Name.Name
	log.Printf("Type:" + tp.Name)
	for _, v := range a.Specs[0].(*ast.TypeSpec).Type.(*ast.StructType).Fields.List {
		tp.Fields[v.Names[0].Name] = &APIField{}
		tp.Fields[v.Names[0].Name].Tags = make(map[string]APIFieldTag)

		switch x := v.Type.(type) {
		case *ast.Ident:
			tp.Fields[v.Names[0].Name].Type = x.Name
		case *ast.ArrayType:
			switch z := x.Elt.(type) {
			case *ast.Ident:
				tp.Fields[v.Names[0].Name].Type = z.Name
				tp.Fields[v.Names[0].Name].Array = true
			case *ast.InterfaceType:
				tp.Fields[v.Names[0].Name].Type = "interface{}"
				tp.Fields[v.Names[0].Name].Array = true

			case *ast.SelectorExpr:
				tp.Fields[v.Names[0].Name].Type = z.X.(*ast.Ident).Name + "." + z.Sel.Name
				tp.Fields[v.Names[0].Name].Array = true
			}

		case *ast.StarExpr:
			switch y := x.X.(type) {
			case *ast.Ident:
				tp.Fields[v.Names[0].Name].Type = y.Name
			case *ast.SelectorExpr:
				switch z := y.X.(type) {
				case *ast.Ident:
					tp.Fields[v.Names[0].Name].Type = z.Name + "." + y.Sel.Name
				}
			}
		case *ast.InterfaceType:
			tp.Fields[v.Names[0].Name].Type = "interface{}"

		case *ast.SelectorExpr:

			switch z := x.X.(type) {
			case *ast.Ident:
				tp.Fields[v.Names[0].Name].Type = z.Name + "." + x.Sel.Name
			}

		case *ast.MapType:

			switch z := x.Value.(type) {
			case *ast.Ident:
				tp.Fields[v.Names[0].Name].Type = ""
				tp.Fields[v.Names[0].Name].Mapkey = x.Key.(*ast.Ident).Name
				tp.Fields[v.Names[0].Name].Mapval = z.Name
				tp.Fields[v.Names[0].Name].Map = true
			case *ast.InterfaceType:
				tp.Fields[v.Names[0].Name].Type = "interface{}"
				tp.Fields[v.Names[0].Name].Array = true
			}

		default:
			log.Printf("%#v", x)
		}
		if v.Tag != nil {
			tgstr := strings.ReplaceAll(v.Tag.Value, "`", "")
			tg, err := structtag.Parse(tgstr)
			if err != nil {
				panic(err)
			}
			for _, tgv := range tg.Keys() {
				atg, err := tg.Get(tgv)
				if err != nil {
					panic(err)
				}
				tp.Fields[v.Names[0].Name].Tags[tgv] = APIFieldTag{
					Key:  tgv,
					Name: atg.Name,
					Opts: atg.Options,
				}
			}
			log.Printf("#%v", tg)
		}

	}
	api.Types[tp.Name] = &tp
}

func addFunction(a *ast.FuncDecl) {
	md := manageComments(a.Doc)

	if md["API"] == "" {
		return
	}
	reqType := &APIParamType{}
	resType := &APIParamType{}

	if len(a.Type.Params.List) > 1 {
		switch x := a.Type.Params.List[1].Type.(type) {
		case *ast.StarExpr:
			reqType.Ispointer = true
			switch y := x.X.(type) {
			case *ast.Ident:
				reqType.Typename = y.Name
			case *ast.SelectorExpr:
				reqType.Typename = y.X.(*ast.Ident).Name + "." + y.Sel.Name
			}
		case *ast.ArrayType:
			reqType.IsArray = true
			switch y := x.Elt.(type) {
			case *ast.Ident:
				reqType.Typename = y.Name
			case *ast.SelectorExpr:
				reqType.Typename = y.X.(*ast.Ident).Name + "." + y.Sel.Name
			case *ast.StarExpr:
				reqType.Ispointer = true
				switch z := y.X.(type) {
				case *ast.Ident:
					reqType.Typename = z.Name
				case *ast.SelectorExpr:
					reqType.Typename = z.X.(*ast.Ident).Name + "." + z.Sel.Name
				}
			}
		case *ast.Ident:
			reqType.Typename = x.Name
		}
	}

	if a.Type.Results != nil && len(a.Type.Results.List) > 0 {

		switch x := a.Type.Results.List[0].Type.(type) {
		case *ast.StarExpr:
			resType.Ispointer = true
			switch y := x.X.(type) {
			case *ast.Ident:
				resType.Typename = y.Name
			case *ast.SelectorExpr:
				resType.Typename = y.X.(*ast.Ident).Name + "." + y.Sel.Name
			}
		case *ast.ArrayType:
			resType.IsArray = true
			switch y := x.Elt.(type) {
			case *ast.Ident:
				resType.Typename = y.Name
			case *ast.SelectorExpr:
				resType.Typename = y.X.(*ast.Ident).Name + "." + y.Sel.Name
			case *ast.StarExpr:
				resType.Ispointer = true
				switch z := y.X.(type) {
				case *ast.Ident:
					resType.Typename = z.Name
				case *ast.SelectorExpr:
					resType.Typename = z.X.(*ast.Ident).Name + "." + z.Sel.Name
				}
			}

		}

		if md["RAW"] == "true" {
			reqType.Typename = md["REQ"]
			resType.Typename = md["RES"]
		}

		verb := md["VERB"]
		if verb == "" {
			verb = http.MethodPost
		}
		fn := APIMethod{
			Desc:    a.Name.Name,
			Verb:    verb,
			Path:    md["PATH"],
			Perm:    md["PERM"],
			ReqType: reqType,
			ResType: resType,
			Raw:     md["RAW"] == "true",
		}
		api.Methods[a.Name.Name] = &fn
	}

}

func load(src string) error {

	api.Types = (make(map[string]*APIType))
	api.Methods = (make(map[string]*APIMethod))

	fset := token.NewFileSet() // positions are relative to fset

	f, err := parser.ParseDir(fset, src, nil, parser.ParseComments)

	if err != nil {
		return err
	}

	for _, v := range f {
		// Print the AST.
		ast.Inspect(v, func(n ast.Node) bool {

			switch x := n.(type) {

			case *ast.GenDecl:
				if x.Tok == token.TYPE {
					addStruct(x)
				} else {
					return true
				}

			case *ast.File:
				c := manageCommentsGroups(x.Comments)
				log.Printf("%+v", c)
			case *ast.FuncDecl:
				addFunction(x)
			case *ast.ValueSpec:
				if x.Names[0].Name == "BASEPATH" {
					api.BasePath = strings.Replace(x.Values[0].(*ast.BasicLit).Value, "\"", "", -1)
				}
				if x.Names[0].Name == "NAMESPACE" {
					api.Namespace = strings.Replace(x.Values[0].(*ast.BasicLit).Value, "\"", "", -1)
				}
				log.Printf("%#v", x)
			case *ast.Package:
				packageName = x.Name
			default:
				//log.Printf("%#v", x)
				return true
			}

			return true
		})
	}

	for k, v := range api.Methods {
		pathmap, ok := httpMapper[v.Path]
		if !ok {
			httpMapper[v.Path] = make(map[string]string)
			pathmap = httpMapper[v.Path]
		}
		pathmap[v.Verb] = k
	}

	return nil
}
